import os.path

class File(str):

    @property
    def NAME(self):
        return File(os.path.basename(self))

    @property
    def DIR(self):
        return File(os.path.dirname(self))

    @property
    def BASEPATH(self):
        return File(os.path.splitext(self)[0])

    @property
    def BASENAME(self):
        return File(os.path.splitext(self.NAME)[0])

    @property
    def EXT(self):
        return os.path.splitext(self)[1]


class Directory(File):
    pass

class MS(Directory):
    pass






import os.path
import importlib
import re
from collections.abc import Sequence

import uuid
from dataclasses import make_dataclass

from omegaconf.omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig
from omegaconf.errors import OmegaConfBaseException
from typing import Any, List, Dict, Optional, Union, Callable

from yaml.error import YAMLError

class ConfigurattError(RuntimeError):
    pass


def _lookup_nameseq(name_seq: List[str], source_dict: Dict):
    """Internal helper: looks up nested item ('a', 'b', 'c') in a nested dict

    Parameters
    ----------
    name_seq : List[str]
        sequence of keys to look up
    source_dict : Dict
        nested dict

    Returns
    -------
    Any
        value if found, else None
    """
    source = source_dict
    names = list(name_seq)
    while names:
        source = source.get(names.pop(0), None)
        if source is None:
            return None
    return source        


def _lookup_name(name: str, *sources: List[Dict]):
    """Internal helper: looks up a nested item ("a.b.c") in a list of dicts

    Parameters
    ----------
    name : str
        section name to look up, e.g. "a.b.c"

    Returns
    -------
    Any
        first matching item found

    Raises
    ------
    NameError
        if matching item is not found
    """
    name_seq = name.split(".")
    for source in sources:
        result = _lookup_nameseq(name_seq, source)
        if result is not None:
            return result
    raise NameError(f"unknown key {name}")


def _flatten_subsections(conf, depth: int = 1, sep: str = "__"):
    """Recursively flattens subsections in a DictConfig (modifying in place)
    A structure such as
        a:
            b: 1
            c: 2
    Becomes
        a__b: 1
        a__c: 2

    Args:
        conf (DictConfig): config to flatten
        depth (int):       depth to which to flatten. Default is 1 level.
        sep (str):         separator to use, default is "__"
    """
    subsections = [(key, value) for key, value in conf.items() if isinstance(value, DictConfig)]
    for name, subsection in subsections:
        conf.pop(name)
        if depth > 1:
            _flatten_subsections(subsection, depth-1, sep)
        for key, value in subsection.items():
            conf[f"{name}{sep}{key}"] = value


def _resolve_config_refs(conf, pathname: str, location: str, name: str, includes: bool, use_sources: Optional[List[DictConfig]], 
                        selfrefs: bool = True, 
                        include_path: Optional[str]=None):
    """Resolves cross-references ("_use" fieds) in config object

    Parameters
    ----------
    conf : OmegaConf object
        input configuration object
    pathname : str
        full path to this confiog (directory component of that is used for _includes)
    location : str
        location of this configuration section, used for messages
    name : str
        name of this configuration file, used for messages
    includes : bool
        If True, "_include" references will be processed
    use_sources : optional list of OmegaConf objects
        one or more config object(s) in which to look up "_use" references. None to disable
    selfrefs (bool, optional): If False, "_use" references will only be looked up in existing config.
        If True (default), they'll also be looked up within the loaded config.
    include_path (str, optional):
        if set, path to each config file will be included in the section as element 'include_path'

    Returns
    -------
    conf : OmegaConf object    
        This may be a new object if a _use key was resolved, or it may be the existing object

    Raises
    ------
    ConfigurattError
        If a _use or _include directive is malformed
    """
    errloc = f"config error at {location or 'top level'} in {name}"

    if isinstance(conf, DictConfig):

        # since _use and _include statements can be nested, keep on processing until all are resolved        
        updated = True
        recurse = 0
        flatten = conf.get("_flatten", 0)
        flatten_sep = conf.get("_flatten_sep", "__")
        for key in "_flatten", "_flatten_sep":
            if key in conf:
                del conf[key]
        
        while updated:
            updated = False
            # check for infinite recursion
            recurse += 1
            if recurse > 20:
                raise ConfigurattError(f"{errloc}: recursion limit exceeded, check your _use and _include statements")

            # handle _include entries
            if includes:
                include_files = conf.get("_include", None)
                if include_files:
                    del conf["_include"]
                    updated = True
                    if isinstance(include_files, str):
                        include_files = [include_files]
                    elif not isinstance(include_files, (tuple, list, ListConfig)) or not all(isinstance(x, str) for x in include_files):
                        raise ConfigurattError(f"{errloc}: _include: must be a string or a list of strings")

                    # load includes
                    accum_incl_conf = OmegaConf.create()
                    for incl in include_files:
                        if not incl:
                            raise ConfigurattError(f"{errloc}: empty _include specifier")

                        # check for (module)filename.yaml style
                        match = re.match("^\\((.+)\\)(.+)$", incl)
                        if match:
                            modulename, filename = match.groups()
                            try:
                                mod = importlib.import_module(modulename)
                            except ImportError as exc:
                                raise ConfigurattError(f"{errloc}: _include {incl}: can't import {modulename} ({exc})")

                            filename = os.path.join(os.path.dirname(mod.__file__), filename)
                            if not os.path.exists(filename):
                                raise ConfigurattError(f"{errloc}: _include {incl}: {filename} does not exist")

                        # absolute path -- one candidate
                        elif os.path.isabs(incl):
                            if not os.path.exists(incl):
                                raise ConfigurattError(f"{errloc}: _include {incl} does not exist")
                            filename = incl
                        # relative path -- scan PATH for candidates
                        else:
                            paths = [os.path.dirname(pathname)] + PATH
                            candidates = [os.path.join(p, incl) for p in paths] 
                            for filename in candidates:
                                if os.path.exists(filename):
                                    break
                            else:
                                raise ConfigurattError(f"{errloc}: _include {incl} not found in {':'.join(PATH)}")

                        # load included file
                        incl_conf = load(filename, location=location, 
                                            name=f"{filename}, included from {name}",
                                            includes=True, 
                                            use_sources=None)   # do not expand _use statements in included files, this is done below

                        if include_path is not None:
                            incl_conf[include_path] = filename

                        # flatten structure
                        if flatten:
                            _flatten_subsections(incl_conf, flatten, flatten_sep)

                        # accumulate included config so that later includes override earlier ones
                        accum_incl_conf = OmegaConf.unsafe_merge(accum_incl_conf, incl_conf)
                    
                    # merge: our section overrides anything that has been included
                    conf = OmegaConf.merge(accum_incl_conf, conf)

            # handle _use entries
            if use_sources is not None:
                merge_sections = conf.get("_use", None)
                if merge_sections:
                    del conf["_use"]
                    updated = True
                    if type(merge_sections) is str:
                        merge_sections = [merge_sections]
                    elif not isinstance(merge_sections, Sequence):
                        raise TypeError(f"invalid {name}._use field of type {type(merge_sections)}")
                    if len(merge_sections):
                        # convert to actual sections
                        merge_sections = [_lookup_name(name, *use_sources) for name in merge_sections]
                        # merge them all together
                        base = merge_sections[0].copy()
                        base.merge_with(*merge_sections[1:])
                        # resolve references before flattening
                        base = _resolve_config_refs(base, pathname=pathname, name=name, 
                                                location=f"{location}._use" if location else "_use", 
                                                includes=includes, 
                                                use_sources=None if use_sources is None else ([conf] + use_sources if selfrefs else use_sources), 
                                                include_path=include_path)
                        if flatten:
                            _flatten_subsections(base, flatten, flatten_sep)
                        base.merge_with(conf)
                        conf = base

        # recurse into content
        for key, value in conf.items_ex(resolve=False):
            if isinstance(value, (DictConfig, ListConfig)):
                value1 = _resolve_config_refs(value, pathname=pathname, name=name, 
                                                location=f"{location}.{key}" if location else key, 
                                                includes=includes, 
                                                use_sources=None if use_sources is None else ([conf] + use_sources if selfrefs else use_sources), 
                                                include_path=include_path)
                # reassigning is expensive, so only do it if there was an actual change 
                if value1 is not value:
                    conf[key] = value1

    # recurse into lists
    elif isinstance(conf, ListConfig):
        # recurse in
        for i, value in enumerate(conf._iter_ex(resolve=False)):
            if isinstance(value, (DictConfig, ListConfig)):
                value1 = _resolve_config_refs(value, pathname=pathname, name=name, 
                                                location=f"{location or ''}[{i}]", 
                                                includes=includes, 
                                                use_sources=None if use_sources is None else ([conf] + use_sources if selfrefs else use_sources), 
                                                include_path=include_path)
                if value1 is not value:
                    conf[i] = value
    return conf


# paths to search for _include statements
PATH = ['.']


def load(path: str, use_sources: Optional[List[DictConfig]] = [], name: Optional[str]=None, location: Optional[str]=None, 
          includes: bool=True, selfrefs: bool=True, include_path: str=None):
    """Loads config file, using a previously loaded config to resolve _use references.

    Args:
        path (str): path to config file
        use_sources (Optional[List[DictConfig]]): list of existing configs to be used to resolve "_use" references, or None to disable
        name (Optional[str]): name of this config file, used for error messages
        location (Optional[str]): location where this config is being loaded (if not at root level)
        includes (bool, optional): If True (default), "_include" references will be processed
        selfrefs (bool, optional): If False, "_use" references will only be looked up in existing config.
            If True (default), they'll also be looked up within the loaded config.
        include_path (str, optional):
            if set, path to each config file will be included in the section as element 'include_path'

    Returns:
        DictConfig: loaded OmegaConf object
    """
    subconf = OmegaConf.load(path)
    name = name or os.path.basename(path)

    return _resolve_config_refs(subconf, pathname=path, location=location, name=name, includes=includes, use_sources=use_sources, include_path=include_path)


def load_nested(filelist: List[str], 
                structured: Optional[DictConfig] = None,
                typeinfo = None,
                use_sources: Optional[List[DictConfig]] = [],
                location: Optional[str] = None,  
                nameattr: Union[Callable, str, None] = None,
                config_class: Optional[str] = None,
                include_path: Optional[str] = None):
    """Builds nested configuration from a set of YAML files corresponding to sub-sections

    Parameters
    ----------
    conf : OmegaConf object
        root OmegaConf object to merge content into
    filelist : List[str]
        list of subsection config files to load
    schema : Optional[DictConfig]
        schema to be applied to each file, if any
    use_sources : Optional[List[DictConfig]]
        list of existing configs to be used to resolve "_use" references, or None to disable
    location : Optional[str]
        if set, contents of files are being loaded under 'location.subsection_name'. If not set, then 'subsection_name' is being
        loaded at root level. This is used for correctly formatting error messages and such.
    nameattr : Union[Callable, str, None]
        if None, subsection_name will be taken from the basename of the file. If set to a string such as 'name', will set 
        subsection_name from that field in the subsection config. If callable, will be called with the subsection config object as a single 
        argument, and must return the subsection name
    config_class : Optional[str]
        name of config dataclass to form (when using typeinfo), if None, then generated automatically
    include_path : Optional[str] 
        if set, path to each config file will be included in the section as element 'include_path'

    Returns
    -------
    dict
        merged config with all subsections

    Raises
    ------
    NameError
        If subsection name is not resolved
    """
    section_content = {} # OmegaConf.create()
    
    for path in filelist:
        # load file
        subconf = load(path, location=location, use_sources=use_sources, include_path=include_path)
        if include_path:
            subconf[include_path] = path

        # figure out section name
        if nameattr is None:
            name = os.path.splitext(os.path.basename(path))[0]
        elif callable(nameattr):
            name = nameattr(subconf) 
        elif nameattr in subconf:
            name = subconf.get(nameattr)
        else:
            raise NameError(f"{path} does not contain a '{nameattr}' field")

        # # resolve _use and _include statements
        # try:
        #     subconf = resolve_config_refs(subconf, f"{location}.{name}" if location else name, conf, subconf))
        # except (OmegaConfBaseException, YAMLError) as exc:
        #     raise ConfigurattError(f"config error in {path}: {exc}")

        # apply schema
        if structured is not None:
            try:
                subconf = OmegaConf.merge(structured, subconf) 
            except (OmegaConfBaseException, YAMLError) as exc:
                raise ConfigurattError(f"schema error in {path}: {exc}")

        section_content[name] = subconf

    if structured is None and typeinfo is not None:
        if config_class is None:
            config_class = "ConfigClass_" + uuid.uuid4().hex
        fields = [(name, typeinfo) for name in section_content.keys()]
        datacls = make_dataclass(config_class, fields)
        structured = OmegaConf.structured(datacls)
        section_content = OmegaConf.merge(structured, section_content)

    return section_content





import os
import logging
import subprocess

from .logging_utils import ConsoleColors, ColorizingFormatter, MultiplexingHandler
from .  import exceptions


def init_logger(name="SCABHA",
           fmt="{asctime}: {message}",
           col_fmt="{asctime}: %s{message}%s"%(ConsoleColors.BEGIN, ConsoleColors.END),
           datefmt="%Y-%m-%d %H:%M:%S", loglevel="INFO"):
    """Returns the global Stimela logger (initializing if not already done so, with the given values)"""
    global log
    if log is None:
        log = logging.getLogger(name)
        log.propagate = False

        level = os.environ.get('SCABHA_LOG_LEVEL') or 'INFO'
        log.setLevel(getattr(logging, level, logging.INFO))

        global log_console_handler, log_formatter, log_boring_formatter, log_colourful_formatter

        # this function checks if the log record corresponds to stdout/stderr output from a cab
        def _is_from_subprocess(rec):
            return hasattr(rec, 'stimela_subprocess_output')

        log_boring_formatter = logging.Formatter(fmt, datefmt, style="{")

        log_colourful_formatter = ColorizingFormatter(col_fmt, datefmt, style="{")

        log_formatter = log_colourful_formatter

        log_console_handler = MultiplexingHandler()
        log_console_handler.setFormatter(log_formatter)
        log_console_handler.setLevel(getattr(logging, loglevel))
        log.addHandler(log_console_handler)

        exceptions.set_logger(log)
    return log


def set_logger(logger):
    global log
    log = logger
    exceptions.set_logger(logger)

def logger():
    return init_logger()

def report_memory():
    """Reports memory status"""
    try:
        output = subprocess.check_output(["/usr/bin/free", "-h"]).decode().splitlines(keepends=False)
    except subprocess.CalledProcessError as exc:
        log.warning(f"/usr/bin/free -h exited with code {exc.returncode}")
        return
    for line in output:
        log.info(line)


def init():
    """Initiatlizes scabha standalone inside container"""
    log.info("Initial memory state:")
    report_memory()


log = None
init_logger()



import sys
import logging

class MultiplexingHandler(logging.Handler):
    """handler to send INFO and below to stdout, everything above to stderr"""
    def __init__(self, info_stream=sys.stdout, err_stream=sys.stderr):
        super(MultiplexingHandler, self).__init__()
        self.info_handler = logging.StreamHandler(info_stream)
        self.err_handler = logging.StreamHandler(err_stream)
        self.multiplex = True

    def emit(self, record):
        handler = self.err_handler if record.levelno > logging.INFO and self.multiplex else self.info_handler
        handler.emit(record)
        # ignore broken pipes, this often happens when cleaning up and exiting
        try:
            handler.flush()
        except BrokenPipeError:
            pass

    def flush(self):
        try:
            self.err_handler.flush()
            self.info_handler.flush()
        except BrokenPipeError:
            pass

    def close(self):
        self.err_handler.close()
        self.info_handler.close()

    def setFormatter(self, fmt):
        self.err_handler.setFormatter(fmt)
        self.info_handler.setFormatter(fmt)

class ConsoleColors():
    WARNING = '\033[93m' if sys.stdin.isatty() else ''
    ERROR   = '\033[91m' if sys.stdin.isatty() else ''
    BOLD    = '\033[1m'  if sys.stdin.isatty() else ''
    DIM     = '\033[2m'  if sys.stdin.isatty() else ''
    GREEN   = '\033[92m' if sys.stdin.isatty() else ''
    ENDC    = '\033[0m'  if sys.stdin.isatty() else ''

    BEGIN = "<COLORIZE>"
    END   = "</COLORIZE>"

    @staticmethod
    def colorize(msg, *styles):
        style = "".join(styles)
        return msg.replace(ConsoleColors.BEGIN, style).replace(ConsoleColors.END, ConsoleColors.ENDC if style else "")

class ColorizingFormatter(logging.Formatter):
    """This Formatter inserts color codes into the string according to severity"""
    def __init__(self, fmt=None, datefmt=None, style="%", default_color=None):
        super(ColorizingFormatter, self).__init__(fmt, datefmt, style)
        self._default_color = default_color or ""

    def format(self, record):
        style = ConsoleColors.BOLD if hasattr(record, 'boldface') else ""
        if hasattr(record, 'color'):
            style += getattr(ConsoleColors, record.color or "None", "")
        elif record.levelno >= logging.ERROR:
            style += ConsoleColors.ERROR
        elif record.levelno >= logging.WARNING:
            style += ConsoleColors.WARNING
        return ConsoleColors.colorize(super(ColorizingFormatter, self).format(record), style or self._default_color)


class SelectiveFormatter(logging.Formatter):
    """Selective formatter. if condition(record) is True, invokes other formatter"""
    def __init__(self, default_formatter, dispatch_list):
        self._dispatch_list = dispatch_list
        self._default_formatter = default_formatter

    def format(self, record):
        for condition, formatter in self._dispatch_list:
            if condition(record):
                return formatter.format(record)
        else:
            return self._default_formatter.format(record)





import os.path, re, stat, itertools, logging, yaml, shlex, importlib
from typing import Any, List, Dict, Optional, Union
from collections import OrderedDict
from enum import Enum, IntEnum
from dataclasses import dataclass
from omegaconf import MISSING, ListConfig, DictConfig

import rich.box
import rich.markup
from rich.table import Table
from rich.markdown import Markdown

import scabha
from scabha import exceptions
from .exceptions import CabValidationError, ParameterValidationError, DefinitionError, SchemaError
from .validate import validate_parameters, Unresolved
from .substitutions import SubstitutionNS
from .basetypes import EmptyDictDefault, EmptyListDefault

## almost supported by omegaconf, see https://github.com/omry/omegaconf/issues/144, for now just use Any
ListOrString = Any   


Conditional = Optional[str]


@dataclass 
class ParameterPolicies(object):
    """This class describes policies that determine how a Parameter is turned into
    cab arguments. Most policies refer to how command-line arguments are formed up,
    although some also apply to Python callable cabs.
    """
    # if true, parameter is passed as key=value, not command line option
    key_value: Optional[bool] = None
    # if true, value is passed as a positional argument, not an option
    positional: Optional[bool] = None
    # if true, value is head-positional, i.e. passed *before* any options
    positional_head: Optional[bool] = None
    # for list-type values, use this as a separator to paste them together into one argument. Otherwise:
    #  * use "list" to pass list-type values as multiple arguments (--option X Y)
    #  * use "[]" to pass list-type values as a list  (--option [X,Y])
    #  * use "repeat" to repeat the option (--option X --option Y)
    repeat: Optional[str] = None
    # prefix for non-positional arguments
    prefix: Optional[str] = None

    # skip this parameter
    skip: Optional[bool] = None
    # if True, implicit parameters will be skipped automatically
    skip_implicits: Optional[bool] = None

    # if set, {}-substitutions on this paramater will not be done
    disable_substitutions: Optional[bool] = None

    # how to pass boolean True values. None = pass option name alone, else pass option name + given value
    explicit_true: Optional[str] = None
    # how to pass boolean False values. None = skip option, else pass option name + given value
    explicit_false: Optional[str] = None

    # if set, a string-type value will be split into a list of arguments using this separator
    split: Optional[str] = None

    # dict of character replacements
    replace: Optional[Dict[str, str]] = None
    
    # Value formatting policies.
    # If set, specifies {}-type format strings used to convert the value(s) to string(s).
    # For a non-list value:
    #   * if 'format_list_scalar' is set, formats the value into a list of strings as fmt[i].format(value, **dict)
    #     example:  ["{0}", "{0}"] will simply repeat the value twice
    #   * if 'format' is set, value is formatted as format.format(value, **dict) 
    # For a list-type value:
    #   * if 'format_list' is set, each element #i formatted separately as fmt[i].format(*value, **dict)
    #     example:  ["{0}", "{2}"] will output elements 0 and 2, and skip element 1 
    #   * if 'format' is set, each element #i is formatted as format.format(value[i], **dict) 
    # **dict contains all parameters passed to a cab, so these can be used in the formatting
    format: Optional[str] = None
    format_list: Optional[List[str]] = None
    format_list_scalar: Optional[List[str]] = None

    # for Python callable cabs: if set, then missing parameters are passed as None values
    # if not set, missing parameters are not passed at all
    pass_missing_as_none: Optional[bool] = None



@dataclass 
class CabManagement:        # defines common cab management behaviours
    environment: Optional[Dict[str, str]] = EmptyDictDefault()
    cleanup: Optional[Dict[str, ListOrString]]     = EmptyDictDefault()   
    wranglers: Optional[Dict[str, ListOrString]]   = EmptyDictDefault()   


# used to classify parameters. Purely for cosmetic and help purposes
ParameterCategory = IntEnum("ParameterCategory", 
                            dict(Required=0, Optional=1, Implicit=2, Obscure=3, Hidden=4),
                            module=__name__)

@dataclass
class Parameter(object):
    """Parameter (of cab or recipe)"""
    info: str = ""
    # for input parameters, this flag indicates a read-write (aka input-output aka mixed-mode) parameter e.g. an MS
    writable: bool = False
    # data type
    dtype: str = "str"
    # specifies that the value is implicitly set inside the step (i.e. not a free parameter). Typically used with filenames 
    implicit: Any = None
    # optonal list of arbitrary tags, used to group parameters
    tags: List[str] = EmptyListDefault()

    # if True, parameter is required
    required: bool = False

    # restrict value choices, i.e. making for an option-type parameter 
    choices:  Optional[List[Any]] = ()

    # for List or Dict-type parameters, restict values of list elements or dict entries to a list of choices
    element_choices: Optional[List[Any]] = None

    # default value
    default: Optional[Any] = None

    # list of aliases for this parameter (i.e. references to other parameters whose schemas/values this parameter shares)
    aliases: Optional[List[str]] = ()

    # if true, treat parameter as a path, and ensure that the parent directories it refers to exist
    mkdir: bool = False
    
    # for file and dir-type parameters: if True, the file(s)/dir(s) must exist. If False, they can be missing.  
    # if None, then the default logic applies: inputs must exist, and outputs don't
    must_exist: Optional[bool] = None

    # if command-line option for underlying binary has a different name, specify it here
    nom_de_guerre: Optional[str] = None

    # policies object, specifying a non-default way to handle this parameter
    policies: ParameterPolicies = ParameterPolicies()

    # Parameter category, purely cosmetic, used for generating help and debug messages. 
    # Assigned automatically if None, but a schema may explicitly mark parameters as e.g. 
    # "obscure" or "hidden"
    category: Optional[ParameterCategory] = None

    # metavar corresponding to this parameter. Used when constructing command-line interfaces
    metavar: Optional[str] = None

    # abbreviated option name for this parameter.  Used when constructing command-line interfaces
    abbreviation: Optional[str] = None

    # arbitrary metadata associated with parameter
    metadata: Dict[str, Any] = EmptyDictDefault() 

    def __post_init__(self):
        def natify(value):
            # convert OmegaConf lists and dicts to native types
            if type(value) in (list, ListConfig):
                return [natify(x) for x in value]
            elif type(value) in (dict, OrderedDict, DictConfig):
                return OrderedDict([(name, natify(value)) for name, value in value.items()])
            return value
        self.default = natify(self.default)
        self.choices = natify(self.choices)

    def get_category(self):
        """Returns category of parameter, auto-setting it if not already preset"""
        if self.category is None:
            if self.required:
                self.category = ParameterCategory.Required
            elif self.implicit is not None:
                self.category = ParameterCategory.Implicit
            else:
                self.category = ParameterCategory.Optional
        return self.category

@dataclass
class Cargo(object):
    name: Optional[str] = None                    # cab name (if None, use image or command name)
    fqname: Optional[str] = None                  # fully-qualified name (recipe_name.step_label.etc.etc.)

    info: Optional[str] = None                    # description
    inputs: Dict[str, Parameter] = EmptyDictDefault()
    outputs: Dict[str, Parameter] = EmptyDictDefault()
    defaults: Dict[str, Any] = EmptyDictDefault()
    backend: Optional[str] = None                 # backend, if not default

    dynamic_schema: Optional[str] = None          # function to call to augment inputs/outputs dynamically

    def __post_init__(self):
        self.fqname = self.fqname or self.name
        self.inputs = OrderedDict((name, Parameter(**schema)) for name, schema in self.inputs.items())
        self.outputs = OrderedDict((name, Parameter(**schema)) for name, schema in self.outputs.items())
        for name in self.inputs.keys():
            if name in self.outputs:
                raise DefinitionError(f"{name} appears in both inputs and outputs")
        self._inputs_outputs = None
        self._implicit_params = set()   # marks implicitly set values
        # pausterized name
        self.name_ = re.sub(r'\W', '_', self.name or "")  # pausterized name
        # config and logger objects
        self.config = self.log = self.logopts = None
        # resolve callable for dynamic schemas
        self._dyn_schema = None
        if self.dynamic_schema is not None:
            if '.' not in self.dynamic_schema:
                raise DefinitionError(f"{self.dynamic_schema}: module_name.function_name expected")
            modulename, funcname = self.dynamic_schema.rsplit(".", 1)
            try:
                mod = importlib.import_module(modulename)
            except ImportError as exc:
                raise DefinitionError(f"can't import {modulename}: {exc}")
            self._dyn_schema = getattr(mod, funcname, None)
            if not callable(self._dyn_schema):
                raise DefinitionError(f"{modulename}.{funcname} is not a valid callable")

    @property
    def inputs_outputs(self):
        if self._inputs_outputs is None:
            self._inputs_outputs = self.inputs.copy()
            self._inputs_outputs.update(**self.outputs)
        return self._inputs_outputs
    
    @property 
    def finalized(self):
        return self.config is not None

    def finalize(self, config=None, log=None, logopts=None, fqname=None, nesting=0):
        if not self.finalized:
            if fqname is not None:
                self.fqname = fqname
            self.config = config
            self.nesting = nesting
            self.log = log
            self.logopts = logopts

    def prevalidate(self, params: Optional[Dict[str, Any]], subst: Optional[SubstitutionNS]=None):
        """Does pre-validation. 
        No parameter substitution is done, but will check for missing params and such.
        A dynamic schema, if defined, is applied at this point."""
        self.finalize()
        # update schemas, if dynamic schema is enabled
        if self._dyn_schema:
            self._inputs_outputs = None
            self.inputs, self.outputs = self._dyn_schema(params, self.inputs, self.outputs)
        # add implicits, if resolved
        for name, schema in self.inputs_outputs.items():
            if schema.implicit is not None and type(schema.implicit) is not Unresolved:
                if name in params and name not in self._implicit_params:
                    raise ParameterValidationError(f"implicit parameter {name} was supplied explicitly")
                if name in self.defaults:
                   raise SchemaError(f"implicit parameter {name} also has a default value")
                params[name] = schema.implicit
                self._implicit_params.add(name)
        # assign unset categories
        for name, schema in self.inputs_outputs.items():
            schema.get_category()

        params = validate_parameters(params, self.inputs_outputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                          check_unknowns=True, check_required=False, check_exist=False,
                                          create_dirs=False, ignore_subst_errors=True)        

        return params

    def validate_inputs(self, params: Dict[str, Any], subst: Optional[SubstitutionNS]=None, loosely=False):
        """Validates inputs.  
        If loosely is True, then doesn't check for required parameters, and doesn't check for files to exist etc.
        This is used when skipping a step.
        """
        assert(self.finalized)
        
        # check inputs
        params1 = validate_parameters(params, self.inputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                                check_unknowns=False, check_required=not loosely, check_exist=not loosely, 
                                                create_dirs=not loosely)
        # check outputs
        params1.update(**validate_parameters(params, self.outputs, defaults=self.defaults, subst=subst, fqname=self.fqname, 
                                                check_unknowns=False, check_required=False, check_exist=False, 
                                                create_dirs=not loosely))
        return params1

    def validate_outputs(self, params: Dict[str, Any], subst: Optional[SubstitutionNS]=None, loosely=False):
        """Validates outputs. Parameter substitution is done. 
        If loosely is True, then doesn't check for required parameters, and doesn't check for files to exist etc.
        """
        assert(self.finalized)
        params.update(**validate_parameters(params, self.outputs, defaults=self.defaults, subst=subst, fqname=self.fqname,
                                                check_unknowns=False, check_required=not loosely, check_exist=not loosely))
        return params

    def make_substitition_namespace(self, params={}):
        from .substitutions import SubstitutionNS
        return SubstitutionNS(**params)

    def rich_help(self, tree, max_category=ParameterCategory.Optional):
        """Generates help into a rich.tree.Tree object"""
        if self.info:
            tree.add("Description:").add(Markdown(self.info))
        # adds tables for inputs and outputs
        for io, title in (self.inputs, "inputs"), (self.outputs, "outputs"):
            for cat in ParameterCategory:
                schemas = [(name, schema) for name, schema in io.items() if schema.get_category() == cat]
                if not schemas:
                    continue
                if cat > max_category:
                    subtree = tree.add(f"[dim]{cat.name} {title}: omitting {len(schemas)}[/dim]")
                    continue
                subtree = tree.add(f"{cat.name} {title}:")
                table = Table.grid("", "", "", padding=(0,2)) # , show_header=False, show_lines=False, box=rich.box.SIMPLE)
                subtree.add(table)            
                for name, schema in schemas: 
                    attrs = []
                    default = self.defaults.get(name, schema.default)
                    if schema.implicit:
                        attrs.append(f"implicit: {schema.implicit}")
                    if default is not None and not isinstance(default, Unresolved):
                        attrs.append(f"default: {default}")
                    if schema.choices:
                        attrs.append(f"choices: {', '.join(schema.choices)}")
                    info = []
                    schema.info and info.append(rich.markup.escape(schema.info))
                    attrs and info.append(f"[dim]\[{rich.markup.escape(', '.join(attrs))}][/dim]")
                    table.add_row(f"[bold]{name}[/bold]", 
                                  f"[dim]{rich.markup.escape(str(schema.dtype))}[/dim]", 
                                  " ".join(info))


ParameterPassingMechanism = Enum("ParameterPassingMechanism", "args yaml", module=__name__)


@dataclass
class Cab(Cargo):
    """Represents a cab i.e. an atomic task in a recipe.
    See dataclass fields below for documentation of fields.

    Additional attributes available after validation with arguments:

        self.input_output:      combined parameter dict (self.input + self.output), maps name to Parameter
        self.missing_params:    dict (name to Parameter) of required parameters that have not been specified
    
    Raises:
        CabValidationError: [description]
    """
    # if set, the cab is run in a container, and this is the image name
    # if not set, commands are run by the native runner
    image: Optional[str] = None                   

    # command to run, inside the container or natively
    command: str = MISSING

    # if set, activates this virtual environment first before running the command (not much sense doing this inside the container)
    virtual_env: Optional[str] = None

    # controls how params are passed. args: via command line argument, yml: via a single yml string
    parameter_passing: ParameterPassingMechanism = ParameterPassingMechanism.args

    # cab management and cleanup definitions
    management: CabManagement = CabManagement()

    # default parameter conversion policies
    policies: ParameterPolicies = ParameterPolicies()

    # copy names of logging levels into wrangler actions
    wrangler_actions =  {attr: value for attr, value in logging.__dict__.items() if attr.upper() == attr and type(value) is int}

    # then add litetal constants for other wrangler actions
    ACTION_SUPPRESS = wrangler_actions["SUPPRESS"] = "SUPPRESS"
    ACTION_DECLARE_SUCCESS = wrangler_actions["DECLARE_SUCCESS"] = "DECLARE_SUPPRESS"
    ACTION_DECLARE_FAILURE = wrangler_actions["DECLARE_FAILURE"] = "DECLARE_FAILURE"


    def __post_init__ (self):
        if self.name is None:
            self.name = self.image or self.command.split()[0]
        Cargo.__post_init__(self)
        for param in self.inputs.keys():
            if param in self.outputs:
                raise CabValidationError(f"cab {self.name}: parameter {param} is both an input and an output, this is not permitted")
        # setup wranglers
        self._wranglers = []
        for match, actions in self.management.wranglers.items():
            replace = None
            if type(actions) is str:
                actions = [actions]
            if type(actions) is not list:
                raise CabValidationError(f"wrangler entry {match}: expected action or list of actions")
            for action in actions:
                if action.startswith("replace:"):
                    replace = action.split(":", 1)[1]
                elif action not in self.wrangler_actions:
                    raise CabValidationError(f"wrangler entry {match}: unknown action '{action}'")
            actions = [self.wrangler_actions[act] for act in actions if act in self.wrangler_actions]
            try:
                rexp = re.compile(match)
            except Exception as exc:
                raise CabValidationError(f"wrangler entry {match} is not a valid regular expression")
            self._wranglers.append((re.compile(match), replace, actions))
        self._runtime_status = None


    def summary(self, params=None, recursive=True, ignore_missing=False):
        lines = [f"cab {self.name}:"] 
        if params is not None:
            for name, value in params.items():
                # if type(value) is validate.Error:
                #     lines.append(f"  {name} = ERR: {value}")
                # else:
                lines.append(f"  {name} = {value}")
            lines += [f"  {name} = ???" for name, schema in self.inputs_outputs.items()
                        if name not in params and (not ignore_missing or schema.required)]
        return lines

    def rich_help(self, tree, max_category=ParameterCategory.Optional):
        tree.add(f"command: {self.command}")
        if self.image:
            tree.add(f"image: {self.image}")
        if self.virtual_env:
            tree.add(f"virtual environment: {self.virtual_env}")
        Cargo.rich_help(self, tree, max_category=max_category)

    def get_schema_policy(self, schema, policy, default=None):
        """Resolves a policy setting. If the policy is set here, returns it. If None and set in the cab,
        returns that. Else returns default value.
        """
        if getattr(schema.policies, policy) is not None:
            return getattr(schema.policies, policy)
        elif getattr(self.policies, policy) is not None:
            return getattr(self.policies, policy)
        else:
            return default

    def build_command_line(self, params: Dict[str, Any], subst: Optional[Dict[str, Any]] = None):
        from .substitutions import substitutions_from

        with substitutions_from(subst, raise_errors=True) as context:
            venv = context.evaluate(self.virtual_env, location=["virtual_env"])
            command = context.evaluate(self.command, location=["command"])

        if venv:
            venv = os.path.expanduser(venv)
            if not os.path.isfile(f"{venv}/bin/activate"):
                raise CabValidationError(f"virtual environment {venv} doesn't exist", log=self.log)
            self.log.debug(f"virtual envirobment is {venv}")
        else:
            venv = None

        command_line = shlex.split(os.path.expanduser(command))
        command = command_line[0]
        args = command_line[1:]
        # collect command
        if "/" not in command:
            from scabha.proc_utils import which
            command0 = command
            command = which(command, extra_paths=venv and [f"{venv}/bin"])
            if command is None:
                raise CabValidationError(f"{command0}: not found", log=self.log)
        else:
            if not os.path.isfile(command) or not os.stat(command).st_mode & stat.S_IXUSR:
                raise CabValidationError(f"{command} doesn't exist or is not executable", log=self.log)

        self.log.debug(f"command is {command}")

        return ([command] + args + self.build_argument_list(params)), venv


    def build_argument_list(self, params):
        """
        Converts command, and current dict of parameters, into a list of command-line arguments.

        pardict:     dict of parameters. If None, pulled from default config.
        positional:  list of positional parameters, if any
        mandatory:   list of mandatory parameters.
        repeat:      How to treat iterable parameter values. If a string (e.g. ","), list values will be passed as one
                    command-line argument, joined by that separator. If True, list values will be passed as
                    multiple repeated command-line options. If None, list values are not allowed.
        repeat_dict: Like repeat, but defines this behaviour per parameter. If supplied, then "repeat" is used
                    as the default for parameters not in repeat_dict.

        Returns list of arguments.
        """

        # collect parameters

        value_dict = dict(**params)

        if self.parameter_passing is ParameterPassingMechanism.yaml:
            return [yaml.safe_dump(value_dict)]

        def get_policy(schema: Parameter, policy: str, default=None):
            return self.get_schema_policy(schema, policy, default)

        def stringify_argument(name, value, schema, option=None):
            key_value = get_policy(schema, 'key_value')
            if key_value:
                return f"{name}={value}"

            if value is None:
                return None
            if schema.dtype == "bool" and not value and get_policy(schema, 'explicit_false') is None:
                return None

            is_list = hasattr(value, '__iter__') and type(value) is not str
            format_policy = get_policy(schema, 'format')
            format_list_policy = get_policy(schema, 'format_list')
            format_scalar_policy = get_policy(schema, 'format_list_scalar')
            split_policy = get_policy(schema, 'split')
            
            if type(value) is str and split_policy:
                value = value.split(split_policy or None)
                is_list = True

            if is_list:
                # apply formatting policies to a list of values
                if format_list_policy:
                    if len(format_list_policy) != len(value):
                        raise CabValidationError("length of format_list_policy does not match length of '{name}'", log=self.log)
                    value = [fmt.format(*value, **value_dict) for fmt in format_list_policy]
                elif format_policy:
                    value = [format_policy.format(x, **value_dict) for x in value]
                else:
                    value = [str(x) for x in value]
            else:
                # apply formatting policies to a scalar valye
                if format_scalar_policy:
                    value = [fmt.format(value, **value_dict) for fmt in format_scalar_policy]
                    is_list = True
                elif format_policy:
                    value = format_policy.format(value, **value_dict)
                else:
                    value = str(value)

            if is_list:
                # check repeat policy and form up representation
                repeat_policy = get_policy(schema, 'repeat')
                if repeat_policy == "list":
                    return [option] + list(value) if option else list(value)
                elif repeat_policy == "[]":
                    val = "[" + ",".join(value) + "]"
                    return [option] + [val] if option else val
                elif repeat_policy == "repeat":
                    return list(itertools.chain([option, x] for x in value)) if option else list(value)
                elif type(repeat_policy) is str:
                    return [option, repeat_policy.join(value)] if option else repeat_policy.join(value)
                elif repeat_policy is None:
                    raise CabValidationError(f"list-type parameter '{name}' does not have a repeat policy set", log=self.log)
                else:
                    raise SchemaError(f"unknown repeat policy '{repeat_policy}'", log=self.log)
            else:
                return [option, value] if option else [value]

        # check for missing parameters and collect positionals

        pos_args = [], []

        for name, schema in self.inputs_outputs.items():
            if schema.required and name not in value_dict:
                raise CabValidationError(f"required parameter '{name}' is missing", log=self.log)
            if name in value_dict:
                positional_first = get_policy(schema, 'positional_head') 
                positional = get_policy(schema, 'positional') or positional_first
                skip = get_policy(schema, 'skip') or (schema.implicit and get_policy(schema, 'skip_implicits', True))
                if positional:
                    if not skip:
                        pargs = pos_args[0 if positional_first else 1]
                        value = stringify_argument(name, value_dict[name], schema)
                        if type(value) is list:
                            pargs += value
                        elif value is not None:
                            pargs.append(value)
                    value_dict.pop(name)

        args = []
                    
        # now check for optional parameters that remain in the dict
        for name, value in value_dict.items():
            if name not in self.inputs_outputs:
                raise RuntimeError(f"unknown parameter '{name}'")
            schema = self.inputs_outputs[name]

            # default behaviour for unset skip_implicits is True
            skip_implicits = get_policy(schema, 'skip_implicits', True)

            if get_policy(schema, 'skip') or (schema.implicit and skip_implicits):
                continue

            key_value = get_policy(schema, 'key_value')

            # apply replacementss
            replacements = get_policy(schema, 'replace')
            if replacements:
                for rep_from, rep_to in replacements.items():
                    try:
                        name = name.replace(rep_from, rep_to)
                    except TypeError:
                        raise TypeError(f"Could not perform policy replacement for parameter [{name}] : {rep_from} => {rep_to}")

            option = (get_policy(schema, 'prefix') or "--") + (schema.nom_de_guerre or name)

            if schema.dtype == "bool":
                if key_value:
                    args += [f"{name}={value}"]
                else:
                    explicit = get_policy(schema, 'explicit_true' if value else 'explicit_false')
                    args += [option, str(explicit)] if explicit is not None else ([option] if value else [])
            else:
                value = stringify_argument(name, value, schema, option=option)
                if type(value) is list:
                    args += value
                elif value is not None:
                    args.append(value)

        return pos_args[0] + args + pos_args[1]


    @property
    def runtime_status(self):
        return self._runtime_status

    def reset_runtime_status(self):
        self._runtime_status = None

    def apply_output_wranglers(self, output, severity):
        suppress = False
        modified_output = output
        for regex, replace, actions in self._wranglers:
            if regex.search(output):
                if replace is not None:
                    modified_output = regex.sub(replace, output)
                for action in actions:
                    if type(action) is int:
                        severity = action
                    elif action is self.ACTION_SUPPRESS:
                        suppress = True
                    elif action is self.ACTION_DECLARE_FAILURE and self._runtime_status is None:
                        self._runtime_status  = False
                        modified_output = "[FAILURE] " + modified_output
                        severity = logging.ERROR
                    elif action is self.ACTION_DECLARE_SUCCESS and self._runtime_status is None:
                        self._runtime_status = True
                        modified_output = "[SUCCESS] " + modified_output
        return (None, 0) if suppress else (modified_output, severity)



@dataclass
class Batch:
    scheduler: str = "slurm"
    cpus: int = 4
    mem: str = "128gb"
    email: Optional[str] = None

    def __init_cab__(self, cab: Cab, params: Dict[str, Any], subst: Optional[Dict[str, Any]], log: Any=None):
        self.cab = cab
        self.log = log
        self.args, self.venv = self.cab.build_command_line(params, subst)




import dataclasses
import os
import os.path
import yaml
import re
from typing import *
from collections import OrderedDict

import pydantic
import pydantic.dataclasses

from omegaconf import OmegaConf, ListConfig, DictConfig

from scabha.basetypes import Unresolved
from .exceptions import Error, ParameterValidationError, SchemaError, SubstitutionErrorList
from .substitutions import SubstitutionNS, substitutions_from
from .types import File, Directory, MS
from .evaluator import Evaluator


def join_quote(values):
    return "'" + "', '".join(values) + "'" if values else ""


def validate_schema(schema: Dict[str, Any]):
    """Checks a set of parameter schemas for internal consistency.

    Args:
        schema (Dict[str, Any]):   dict of parameter schemas

    Raises:
        SchemaError: [description]
    """

    pass


def is_file_type(dtype):
    return dtype in (File, Directory, MS)

def is_filelist_type(dtype):
    return dtype in (List[File], List[Directory], List[MS])


def evaluate_and_substitute(inputs: Dict[str, Any], 
                            subst: SubstitutionNS, 
                            corresponding_ns: SubstitutionNS,
                            defaults: Dict[str, Any] = {},
                            ignore_subst_errors: bool = False, 
                            location: List[str] = []):
    with substitutions_from(subst, raise_errors=True) as context:
        evaltor = Evaluator(subst, context, location=location)
        inputs = evaltor.evaluate_dict(inputs, corresponding_ns=corresponding_ns, defaults=defaults,
                                        raise_substitution_errors=False)
        # collect errors
        if not ignore_subst_errors:
            errors = []
            for value in inputs.values():
                if type(value) is Unresolved:
                    errors += value.errors
            # check for substitution errors
            if errors:
                raise SubstitutionErrorList(*errors)
    return inputs


def validate_parameters(params: Dict[str, Any], schemas: Dict[str, Any], 
                        defaults: Optional[Dict[str, Any]] = None,
                        subst: Optional[SubstitutionNS] = None,
                        fqname: str = "",
                        check_unknowns=True,    
                        check_required=True,
                        check_exist=True,
                        create_dirs=False,
                        ignore_subst_errors=False
                        ) -> Dict[str, Any]:
    """Validates a dict of parameter values against a given schema 

    Args:
        params (Dict[str, Any]):   map of input parameter values
        schemas (Dict[str, Any]):   map of parameter names to schemas. Each schema must contain a dtype field and a choices field.
        defaults (Dict[str, Any], optional): dictionary of default values to be used when a value is missing
        subst (SubsititionNS, optional): namespace to do {}-substitutions on parameter values
        fqname: fully-qualified name of the parameter set (e.g. "recipe_name.step_name"), used in error messages. If not given,
                errors will report parameter names only

        check_unknowns (bool): if True, unknown parameters (not in schema) raise an error
        check_required (bool): if True, missing parameters with required=True will raise an error
        check_exist (bool): if True, files with must_exist={None,True} in schema must exist, or will raise an error. 
                            If False, only files with must_exist=True must exist.
        create_dirs (bool): if True, non-existing directories in filenames (and parameters with mkdir=True in schema) 
                            will be created.
        ignore_subst_errors (bool): if True, substitution errors will be ignored


    Raises:
        ParameterValidationError: parameter fails validation
        SchemaError: bad schema
        SubstitutionErrorList: list of substitution errors, if they occur

    Returns:
        Dict[str, Any]: validated dict of parameters

    TODO:
        add options to propagate all errors out (as values of type Error) in place of exceptions?
    """
    # define function for converting parameter name into "fully-qualified" name
    if fqname:
        mkname = lambda name: f"{fqname}.{name}"
    else:
        mkname = lambda name: name

    # check for unknowns
    if check_unknowns:
        for name in params:
            if name not in schemas:
                raise ParameterValidationError(f"unknown parameter '{mkname(name)}'")
    
    # only evaluate the subset for which we have schemas
    inputs = OrderedDict((name, value) for name, value in params.items() if name in schemas)
    
    # build dict of all defaults 
    all_defaults = {name: schema.default for name, schema in schemas.items() if schema.default is not None}
    if defaults:
        all_defaults.update(**{name: value for name, value in defaults.items() if name in schemas})

    # update missing inputs from defaults
    inputs.update(**{name: value for name, value in all_defaults.items() if name not in inputs})

    # perform substitution
    if subst is not None:
        inputs = evaluate_and_substitute(inputs, subst, subst.current, defaults=all_defaults, 
                                        ignore_subst_errors=ignore_subst_errors, location=[fqname])

    # split inputs into unresolved substitutions, and proper inputs
    unresolved = {name: value for name, value in inputs.items() if isinstance(value, Unresolved)}
    inputs = {name: value for name, value in inputs.items() if not isinstance(value, Unresolved)}

    # check that required args are present
    if check_required:
        missing = [mkname(name) for name, schema in schemas.items() 
                    if schema.required and inputs.get(name) is None and name not in unresolved]
        if missing:
            raise ParameterValidationError(f"missing required parameters: {join_quote(missing)}")

    # create dataclass from parameter schema
    validated = {}
    dtypes = {}
    fields = []

    # maps parameter names to/from field names. Fields have "_" not "-"
    name2field = {}
    field2name = {}

    for name, schema in schemas.items():
        value = inputs.get(name)
        if value is not None:
            try:
                dtypes[name] = dtype_impl = eval(schema.dtype, globals())
            except Exception as exc:
                raise SchemaError(f"invalid {mkname(name)}.dtype = {schema.dtype} ({exc})")

            # sanitize name: dataclass won't take hyphens or periods
            fldname = re.sub("\W", "_", name)
            while fldname in field2name:
                fldname += "_"
            field2name[fldname] = name
            name2field[name] = fldname

            fields.append((fldname, dtype_impl))
            
            # OmegaConf dicts/lists need to be converted to standard containers for pydantic to take them
            if isinstance(value, (ListConfig, DictConfig)):
                inputs[name] = OmegaConf.to_container(value)

    dcls = dataclasses.make_dataclass("Parameters", fields)

    # convert this to a pydantic dataclass which does validation
    pcls = pydantic.dataclasses.dataclass(dcls)

    # check Files etc. and expand globs
    for name, value in list(inputs.items()):
        # get schema from those that need validation, skip if not in schemas
        schema = schemas.get(name)
        if schema is None:
            continue
        # skip errors
        if value is None or isinstance(value, Error):
            continue
        dtype = dtypes[name]

        is_file = is_file_type(dtype)
        is_file_list = is_filelist_type(dtype)

        # must this file exist? Schema may force this check, otherwise follow the default check_exist policy
        must_exist = check_exist if schema.must_exist is None else schema.must_exist

        if is_file or is_file_list:
            # match to existing file(s)
            if type(value) is str:
                # try to interpret string as a formatted list (a list substituted in would come out like that)
                try:
                    files = yaml.safe_load(value)
                    if type(files) is not list:
                        files = [value]
                except Exception as exc:
                    files = [value]
            elif isinstance(value, (list, tuple)):
                files = value
            else:
                raise ParameterValidationError(f"'{mkname(name)}={value}': invalid type '{type(value)}'")
          
            # check for existence of all files in list, if needed
            if must_exist: 
                not_exists = [f for f in files if not os.path.exists(f)]
                if not_exists:
                    raise ParameterValidationError(f"'{mkname(name)}': {','.join(not_exists)} doesn't exist")

            # check for single file/dir
            if dtype in (File, Directory, MS):
                if len(files) > 1:
                    raise ParameterValidationError(f"'{mkname(name)}': multiple files given ({value})")
                # check that files are files and dirs are dirs
                if os.path.exists(files[0]):
                    if dtype is File:
                        if not os.path.isfile(files[0]):
                            raise ParameterValidationError(f"'{mkname(name)}': {value} is not a regular file")
                    else:
                        if not os.path.isdir(files[0]):
                            raise ParameterValidationError(f"'{mkname(name)}': {value} is not a directory")
                inputs[name] = files[0]
                if create_dirs:
                    dirname = os.path.dirname(files[0])
                    if dirname:
                        os.makedirs(dirname, exist_ok=True)
            # else make list
            else:
                # check that files are files and dirs are dirs
                if dtype is List[File]:
                    if not all(os.path.isfile(f) for f in files if os.path.exists(f)):
                        raise ParameterValidationError(f"{mkname(name)}: {value} matches non-files")
                else:
                    if not all(os.path.isdir(f) for f in files if os.path.exists(f)):
                        raise ParameterValidationError(f"{mkname(name)}: {value} matches non-directories")
                inputs[name] = files
                if create_dirs:
                    for path in files:
                        dirname = os.path.dirname(path)
                        if dirname:
                            os.makedirs(dirname, exist_ok=True)

    # validate
    try:   
        validated = pcls(**{name2field[name]: value for name, value in inputs.items() if name in schemas and value is not None})
    except pydantic.ValidationError as exc:
        errors = [f"'{'.'.join(map(str, err['loc']))}': {err['msg']}" for err in exc.errors()]
        raise ParameterValidationError(', '.join(errors))

    validated = {field2name[fld]: value for fld, value in dataclasses.asdict(validated).items()}

    # check choice-type parameters
    for name, value in validated.items():
        schema = schemas[name]
        if schema.choices and value not in schema.choices:
            raise ParameterValidationError(f"{mkname(name)}: invalid value '{value}'")

    # check for mkdir directives
    if create_dirs:
        for name, value in validated.items():
            if schemas[name].mkdir:
                dirname = os.path.dirname(value)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)

    # add in unresolved values
    validated.update(**unresolved)

    return validated



from typing import List

logger = None

def set_logger(log):
    global logger
    logger = log


class Error(str):
    """A string that's marked as an error"""
    pass


class ScabhaBaseException(Exception):
    def __init__(self, message, log=None):
        Exception.__init__(self, message)
        if log is not None:
            if not hasattr(log, 'error'):
                log = logger
            if log is not None:
                log.error(message)
        self.logged = log is not None

class SchemaError(ScabhaBaseException):
    pass

class DefinitionError(ScabhaBaseException):
    pass

class StepValidationError(ScabhaBaseException):
    pass

class CabValidationError(ScabhaBaseException):
    pass

class ParameterValidationError(ScabhaBaseException):
    pass

class SubstitutionError(ScabhaBaseException):
    pass

class UnsetError(ScabhaBaseException):
    def __init__(self, message, element, log=None):
        super().__init__(message, log)
        self.element = element

class ParserError(ScabhaBaseException):
    pass

class FormulaError(ScabhaBaseException):
    pass

class CyclicSubstitutionError(SubstitutionError):
    def __init__(self, location: List[str], other_location: List[str]):
        self.location = ".".join(location)
        self.other_location = ".".join(other_location)
        super().__init__(f"'{{{self.location}}}' is a cyclic substition")

class SubstitutionErrorList(ScabhaBaseException):
    def __init__(self, *errors):
        self.errors = errors
        super().__init__(f"{len(errors)} substitution error(s): {'; '.join(map(str, errors))}")



from dataclasses import field, dataclass
from collections import OrderedDict
from typing import List


def EmptyDictDefault():
    return field(default_factory=lambda:OrderedDict())


def EmptyListDefault():
    return field(default_factory=lambda:[])
    

@dataclass
class Unresolved(object):
    value: str = ""
    errors: List[Exception] = EmptyListDefault

    def __post_init__(self):
        if not self.value:
            self.value = "; ".join(map(str, self.errors))

    def __str__(self):
        return f"Unresolved({self.value})"



#!/usr/bin/env python
from setuptools import setup

requirements = ["pyyaml",
                "nose>=1.3.7",
                "omegaconf",
                "click",
                "pydantic",
                "ruamel.yaml",
                "pyparsing",
                "pytest",
                "rich"
                ],

PACKAGE_NAME = "scabha"

__version__ = "2.0beta1"

setup(name=PACKAGE_NAME,
      version=__version__,
      description="Onboard services for passengers of Stimela (https://github.com/caracal-pipeline/stimela2) cabs",
      author="Oleg Smirnov & RATT",
      author_email="osmirnov@gmail.com",
      url="https://github.com/caracal-pipeline/scabha2",
      packages=["scabha"],
      package_data={"scabha": []},
      install_requires=requirements,
      scripts=[],
      classifiers=[],
      )



import re
from collections import OrderedDict
from scabha.cargo import EmptyDictDefault
from dataclasses import dataclass
from contextlib import contextmanager
from typing import Any, Dict, Optional, Union

from .validate import Error
from .exceptions import SubstitutionError


# thanks to https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729
def multireplace(string, replacements, ignore_case=False):
    """
    Given a string and a replacement map, it returns the replaced string.
    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {value to find: value to replace}
    :param bool ignore_case: whether the match should be case insensitive
    :rtype: str
    """
    # If case insensitive, we need to normalize the old string so that later a replacement
    # can be found. For instance with {"HEY": "lol"} we should match and find a replacement for "hey",
    # "HEY", "hEy", etc.
    if ignore_case:
        def normalize_old(s):
            return s.lower()

        re_mode = re.IGNORECASE

    else:
        def normalize_old(s):
            return s

        re_mode = 0

    replacements = {normalize_old(key): val for key, val in replacements.items()}
    
    # Place longer ones first to keep shorter substrings from matching where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    rep_sorted = sorted(replacements, key=len, reverse=True)
    rep_escaped = map(re.escape, rep_sorted)
    
    # Create a big OR regex that matches any of the substrings to replace
    pattern = re.compile("|".join(rep_escaped), re_mode)
    
    # For each match, look up the new string in the replacements, being the key the normalized old string
    return pattern.sub(lambda match: replacements[normalize_old(match.group(0))], string)

def _fqname(parent_names, name):
    return '.'.join((parent_names or []) + [name])


@contextmanager
def forgive_errors(error_types={AttributeError}, mode=""):
    previous = SubstitutionNamespace._forgiving_.copy()
    for err in error_types:
        SubstitutionNamespace._forgiving_[err] = mode
    try:
        yield True
    finally:
        SubstitutionNamespace._forgiving_ = previous

def forgive_most_errors(mode=""):
    return forgive_errors({AttributeError, TypeError, ValueError}, mode="")


class SubstitutionNamespace(OrderedDict):
    """Implements a namespace that can do {}-substitutions on itself
    """
    @dataclass
    class Properties(object):
        mutable: bool = True
        updated: bool = False
        forgiving: dict = EmptyDictDefault

    _default_prop_ = Properties()

    # forgiving mode for various errors. If False/None, raise AttributeError. If True, return "{attr}"" when attribute is not known. If string, return given string
    _forgiving_ = {}

    def __init__(self, **kw):
        """Initializes the namespace. Keywords are _add_'ed as items in the namespace
        """
        super().__setattr__('_props_', SubstitutionNamespace.Properties())
        super().__setattr__('_child_props_', {})
        super().__setattr__('_forgave_', set())
        SubstitutionNamespace._update_(self, **kw)

    def copy(self):
        newcopy = SubstitutionNamespace()
        OrderedDict.__setattr__(newcopy, '_props_', self._props_)
        OrderedDict.__setattr__(newcopy, '_child_props_', self._child_props_.copy())
        OrderedDict.__setattr__(newcopy, '_forgave_', self._forgave_.copy())
        for key, value in self.items():
            OrderedDict.__setitem__(newcopy, key, value)
        return newcopy

    def _update_(self, **kw):
        """Updates items in the namespace using _add_()
        """
        for name, value in kw.items():
            SubstitutionNamespace._add_(self, name, value)

    def _merge_(self, ns):
        """Recursively merges in one namespace into another
        """
        for name, value in ns.items():
            if name not in self:
                SubstitutionNamespace._add_(self, name, value)
            else:
                old_value = super().get(name)
                if isinstance(old_value, SubstitutionNamespace) and \
                    isinstance(value, (dict, OrderedDict, SubstitutionNamespace)):
                    old_value._merge_(**value)
                else:
                    SubstitutionNamespace._add_(self, name, value)

    def _add_(self, k: str, v: Any, forgiving={}, mutable=True):
        """Adds an item to the namespace.

        Args:
            k (str): item key
            v (Any): item value. A dict or OrderedDict value becomes a SubstitutionNamespace automatically
            forgiving (bool, optional): If True, sub-namespace is "forgiving" with references to missing items,
                returning "(name)" for ns.name if name is missing. If False, such references result in an AttributeError.
                Default is False.
            mutable (bool, optional): If False, sub-namespace is immutable and not will not have substitutions done inside it. Defaults to True.
        """
        if forgiving is True:
            forgiving = {AttributeError: True}
        props = SubstitutionNamespace.Properties(mutable=mutable, forgiving=forgiving)
        if type(v) in (dict, OrderedDict):
            v = SubstitutionNamespace(**v)
        if isinstance(v, SubstitutionNamespace):
            OrderedDict.__setattr__(v, '_props_', props)
        self._child_props_[k] = props
        super().__setitem__(k, v)

    def __setattr__(self, name: str, value: Any) -> None:
        SubstitutionNamespace._add_(self, name, value)

    def __setitem__(self, k: str, v: Any) -> None:
        SubstitutionNamespace._add_(self, k, v)

    def __forgiving_mode__ (self, err):
        forgive = SubstitutionNamespace._forgiving_.get(err) 
        if forgive is None:
            forgive = self._props_.forgiving.get(err)
        return forgive

    def __getattr__(self, name: str) -> Any:
        if name in self:
            return super().get(name)
        else:
            # if global mode is set, overrides local mode
            forgive = self.__forgiving_mode__(AttributeError)
            # if string, or True, return forgive-value
            if type(forgive) is str or forgive:
                self._forgave_.add(name)
                return forgive if type(forgive) is str else f"(name)"  
            else:
                raise AttributeError(name)

    def _substitute_(self, subst: Optional['SubstitutionNamespace'] = None, parent_names=[]):
        """Recursively substitutes {}-strings within this namespace

        Args:
            subst (SubstitutionNamespace, optional): Namespace used to look up substitutions. Defaults to self.
            parent_names (list): list of parent names (including name of this namespace, if any)

        Returns:
            SubstitutionNamespace, updated, unresolved: output namespace (same as self if copy=False), count of updates, list of unresolved substitutions
        """
        updated = 0
        unresolved = []
        output = self
        # loop over parameters and find ones to substitute
        for name, value in super().items():
            props = self._child_props_[name]
            updated1 = 0
            unresolved1 = []
            # substitute strings
            if isinstance(value, str) and not isinstance(value, Error) and "{" in value:
                # format string value
                try:
                    # protect "{{" and "}}" from getting converted to a single brace by pre-replacing them
                    newvalue = multireplace(value, {'{{': '\u00AB', '}}': '\u00BB'})
                    newvalue = newvalue.format(**(subst or output))
                    newvalue = multireplace(newvalue, {'\u00AB': '{{', '\u00BB': '}}'})
                    updated1 = int(value != newvalue)
                except Exception as exc:
                    forgive = self.__forgiving_mode__(type(exc))
                    if type(forgive) is str:
                        newvalue = forgive
                        updated1 = int(value != newvalue)
                    elif forgive:
                        newvalue = "(name)"
                        updated1 = int(value != newvalue)
                    else:
                        location = f"{'.'.join(parent_names + [name])}='{value}'"
                        if type(exc) is AttributeError:
                            err = SubstitutionError(f"'{{{' '.join(exc.args)}}}' unresolved in {location}")
                        else:
                            err = SubstitutionError(f"{exc} in {location} ({type(exc)})")
                        newvalue = err
                        unresolved1 = [err] 
                        updated1 = 1
            # else substitute into mutable sub-namespaces
            elif isinstance(value, SubstitutionNamespace) and props.mutable:
                newvalue, updated1, unresolved1 = value._substitute_(subst or output, parent_names=parent_names + [name])
            elif isinstance(value, Exception):
                unresolved1 = [value]
            # has something changed? make copy of ourselves if so
            if updated1:
                if output is self:
                    output = self.copy()
                OrderedDict.__setitem__(output, name, newvalue)
            # update counters
            updated += updated1
            unresolved += unresolved1

        return output, updated, unresolved

    def _clear_forgivens_(self):
        super().__setattr__('_forgave_', set())
        for child in self.values():
            if isinstance(child, SubstitutionNamespace):
                child._clear_forgivens_()

    def _collect_forgivens_(self, name: Optional[str] = None):
        own_name = name or "."
        result = [f"{own_name}.{key}" for key in self._forgave_]
        for child_name, child in self.items():
            if isinstance(child, SubstitutionNamespace):
                result += child._collect_forgivens_(f"{name}.{child_name}" if name is not None else child_name)
        return result

    def _finalize_braces_(self):
        output = self
        for name, value in self.items():
            props = self._child_props_[name]
            updated = False
            if isinstance(value, SubstitutionNamespace) and props.mutable:
                newvalue = value._finalize_braces_()
                updated = newvalue is not value
            elif isinstance(value, str):
                newvalue = value.format()  # this converts {{ and }} to { and }
                updated = newvalue != value
            if updated:
                if output is self:
                    output = self.copy()
                OrderedDict.__setitem__(output, name, newvalue)
        return output

    def _print_(self, prefix="", printfunc=print):
        for name, value in self.items():
            if name.startswith("_") or name.endswith("_"):
                continue
            if isinstance(value, SubstitutionNamespace):
                printfunc(f"{prefix}{name}:")
                value._print_(prefix + "  ")
            elif isinstance(value, Exception):
                printfunc(f"{prefix}{name}: ERR: {value}")
            else:
                printfunc(f"{prefix}{name}: {value}")


def self_substitute(ns: SubstitutionNamespace, name: Optional[str] = None, debugprint = None):
    """Resolves {}-substitutions within a namespace.

    Args:
        ns (SubstitutionNamespace): namespace to do substitutions in.
        name (Optional[str], optional): name of this namespace, used in messages.
        debugprint (callable, optional): if set, function used to print debug messages.

    Raises:
        SubstitutionError: [description]

    Returns:
        SubstitutionNamespace: copy of namespace with substitutions in it. Will be the same as ns if no substitutions done
    """
    ns._clear_forgivens_()

    debugprint and debugprint("--- before substitution ---")
    debugprint and ns._print_(printfunc=debugprint, prefix="  ")

    # repeat as long as values keep changing, but qut after 10 cycles in case of infinite cross-refs
    for i in range(10):
        ns, updated, unresolved = ns._substitute_()
        debugprint and debugprint(f"--- iteration {i} updated {updated} unresolved {unresolved} ---")
        if not updated:
            break 
        debugprint and ns._print_(printfunc=debugprint, prefix="  ")
    else:
        raise SubstitutionError("recursion limit exceeded while evaluating {}-substitutions. This is usally caused by cyclic (cross-)substitutions.")

    # clear up "{{"s
    debugprint and debugprint(f"--- finalizing curly braces ---")
    ns = ns._finalize_braces_()
    debugprint and ns._print_(printfunc=debugprint, prefix="  ")

    return ns, unresolved, ns._collect_forgivens_(name)


# def copy_updates(src: SubstitutionNamespace, dest: Dict[str, Any]):
#     for name, value in src.items():
#         props = src._child_props_[name]
#         if props.updated:
#             dest[name] = value



import glob

import pyparsing
pyparsing.ParserElement.enable_packrat()
from pyparsing import *
from pyparsing import common

from .substitutions import SubstitutionError, SubstitutionContext
from .basetypes import Unresolved
from .exceptions import *

import typing
from typing import Dict, List, Any


_parser = None

def construct_parser():
    lparen = Literal("(").suppress()
    rparen = Literal(")").suppress()
    comma = Literal(",").suppress()
    period = Literal(".").suppress()
    string = (QuotedString('"') | QuotedString("'"))("constant")

    bool_false = (Keyword("False") | Keyword("false"))("bool_false").set_parse_action(lambda:[False])
    bool_true = (Keyword("True") | Keyword("true"))("bool_true").set_parse_action(lambda:[True])
    boolean = (bool_true | bool_false)("constant")

    number = common.number("constant")

    # identifier = Word(alphas, alphanums + "_")
    fieldname = Word(alphas + "_", alphanums + "_-@")
    nested_field = Group(fieldname + ZeroOrMore(period + fieldname).leave_whitespace()).leave_whitespace()("namespace_lookup")

    anyseq = CharsNotIn(",)")("constant")

    UNSET = Keyword("UNSET")("unset")
    # functions
    IF = Keyword("IF")
    IFSET = Keyword("IFSET")
    GLOB = Keyword("GLOB")
    EXISTS = Keyword("EXISTS")

    # allow expression to be used recursively
    expr = Forward()

    # parenthesized expression
    subexpr = Group(lparen + expr + rparen)("subexpression")

    atomic_value = (boolean | UNSET | nested_field | string | number) #("atomic_value")
    comma_empty = Group(comma + Empty())("empty")
    varg = expr
    comma_varg = comma + varg

    if_ = IF + lparen + varg + comma_varg + comma_varg + Optional(comma_varg) + rparen
    ifset_ = IFSET + lparen + nested_field + Optional(comma_varg|comma_empty) + \
                     Optional(comma_varg|comma_empty) + Optional(comma_varg|comma_empty) + rparen
    glob_ = GLOB + lparen + (varg|anyseq) + rparen
    exists_ = EXISTS + lparen + (varg|anyseq) + rparen

    # function call
    function = (ifset_ | if_ | glob_  | exists_)("function")

    operators = (
        [(Literal("**")("op2"), 1, opAssoc.LEFT)] + 
        [(Literal(x)("op1"), 1, opAssoc.RIGHT) for x in "-+~"] + 
        [(Literal(x)("op2"), 2, opAssoc.LEFT) for x in ("*", "//", "/", "%")] +
        [(Literal(x)("op2"), 2, opAssoc.LEFT) for x in "+-"] +
        [(Literal(x)("op2"), 2, opAssoc.LEFT) for x in ("<<", ">>")] +
        [
            (Literal("&")("op2"), 2, opAssoc.LEFT),
            (Literal("^")("op2"), 2, opAssoc.LEFT),
            (Literal("|")("op2"), 2, opAssoc.LEFT)
        ] +
        [(Literal(x)("op2"), 2, opAssoc.LEFT) for x in ("==", "!=", ">", "<", ">=", "<=")] +
        [(Keyword(x)("op2"), 2, opAssoc.LEFT) for x in ("in", "not in")] +
        [   (Keyword("not")("op1"), 1, opAssoc.RIGHT),
            (Keyword("and")("op2"), 2, opAssoc.LEFT),
            (Keyword("or")("op2"), 2, opAssoc.LEFT)
        ]
    )
    infix = infix_notation(function | atomic_value, operators)("subexpression")

    expr <<= function | infix

    return expr


if _parser is None:
    _parser = construct_parser()

_parse_cache = {}

def parse_string(text: str, location: List[str] = []):
    """parses a formula, returns ParseResults object. Implements cache.

    Args:
        text (str): formula to parse
        location (List[str], optional): Hierarchical location, used for erro messages
        
    Raises:
        ParserError: on any parse error

    Returns:
        ParseResults: on successful parse
    """
    parse_results = _parse_cache.get(text)
    if parse_results is None:
        try:
            parse_results = _parser.parse_string(text, parse_all=True)
        except Exception as exc:
            parse_results = ParserError(f"{'.'.join(location)}: error parsing formula ({exc})")
        _parse_cache[text] = parse_results

    if isinstance(parse_results, Exception):
        raise parse_results

    return parse_results

def is_empty(result):
    return result is None or (type(result) is ParseResults and result._name == "empty")

class UNSET(object):
    def __init__(self, name) -> None:
        self.name = name

class Evaluator(object):
    def __init__(self,  ns: Dict[str, Any], 
                        subst_context: typing.Optional[SubstitutionContext] = None, 
                        location: List[str] = []):
        self.ns = ns
        self.subst_context = subst_context
        self.location = location

        self._UNARY_OPERATORS = {
            '+':    lambda x: +self._evaluate_result(x),
            '-':    lambda x: -self._evaluate_result(x),
            '~':    lambda x: ~self._evaluate_result(x),
            'not':  lambda x: ~self._evaluate_result(x)
        }
        self._BINARY_OPERATORS = {
            '**':   lambda x,y: self._evaluate_result(x) ** self._evaluate_result(y),
            '*':   lambda x,y: self._evaluate_result(x) * self._evaluate_result(y),
            '/':   lambda x,y: self._evaluate_result(x) / self._evaluate_result(y),
            '//':   lambda x,y: self._evaluate_result(x) // self._evaluate_result(y),
            '+':   lambda x,y: self._evaluate_result(x) + self._evaluate_result(y),
            '-':   lambda x,y: self._evaluate_result(x) - self._evaluate_result(y),
            '<<':   lambda x,y: self._evaluate_result(x) << self._evaluate_result(y),
            '>>':   lambda x,y: self._evaluate_result(x) >> self._evaluate_result(y),
            '&':   lambda x,y: self._evaluate_result(x) & self._evaluate_result(y),
            '^':   lambda x,y: self._evaluate_result(x) ^ self._evaluate_result(y),
            '|':   lambda x,y: self._evaluate_result(x) | self._evaluate_result(y),
            '==':   lambda x,y: self._evaluate_result(x) == self._evaluate_result(y),
            '!=':   lambda x,y: self._evaluate_result(x) != self._evaluate_result(y),
            '<=':   lambda x,y: self._evaluate_result(x) <= self._evaluate_result(y),
            '<':   lambda x,y: self._evaluate_result(x) < self._evaluate_result(y),
            '>=':   lambda x,y: self._evaluate_result(x) >= self._evaluate_result(y),
            '>':   lambda x,y: self._evaluate_result(x) > self._evaluate_result(y),
            # 'is':   lambda x,y: self._evaluate_result(x) is self._evaluate_result(y),
            # 'is not':   lambda x,y: self._evaluate_result(x) is not self._evaluate_result(y),
            'in':   lambda x,y: self._evaluate_result(x) in self._evaluate_result(y),
            'not in':   lambda x,y: self._evaluate_result(x) not in self._evaluate_result(y),
            'and':   lambda x,y: self._evaluate_result(x) and self._evaluate_result(y),
            'or':   lambda x,y: self._evaluate_result(x) or self._evaluate_result(y),
        }        

    def _resolve(self, value):
        if type(value) is str and self.subst_context is not None:
            try:
                value = self.subst_context.evaluate(value, location=self.location)
            except Exception as exc:
                raise SubstitutionError(f"{value}: {exc}")
        return value

    def empty(self, *args):
        return None

    def unset(self, *args):
        return UNSET

    def constant(self, value):
        return self._resolve(value)

    def subexpression(self, value):
        return self._evaluate_result(value)

    def namespace_lookup(self, *args):
        if len(args) == 1 and type(args[0]) is ParseResults:
            args = args[0]
            assert args._name == "namespace_lookup"
        value = self.ns
        fields = list(args)
        while fields:
            fld = fields.pop(0)
            # last element allowed to be UNSET, otherwise substitution error
            if fld not in value:
                if fields:
                    raise SubstitutionError(f"{'.'.join(self.location)}: '{'.'.join(args)}' is not defined (at '{fld}')")
                else:
                    return UNSET('.'.join(args))
            value = value[fld]
        return self._resolve(value)

    def function(self, funcname, *args):
        method = f"func_{funcname}"
        if not hasattr(self, method):
            raise NameError(f"{'.'.join(self.location)}: unknown function '{funcname}'")
        return getattr(self, method)(*args)

    def func_IF(self, *args):
        if len(args) < 3 or len(args) > 4:
            raise FormulaError(f"{'.'.join(self.location)}: IF() expects 3 or 4 arguments, got {len(args)}")
        conditional, if_true, if_false = args[:3]
        if_unset = args[3] if len(args) == 4 else None

        cond = self._evaluate_result(conditional, allow_unset=if_unset is not None)
        if type(cond) is UNSET:
            if if_unset is None:
                raise SubstitutionError(f"{'.'.join(self.location)}: '{cond.name}' is not defined")
            result = if_unset
        else:
            result = if_true if cond else if_false    
            
        return self._evaluate_result(result)

    def func_IFSET(self, *args):
        if len(args) < 1 or len(args) > 3:
            raise FormulaError(f"{'.'.join(self.location)}: IFSET() expects 1 to 3 arguments, got {len(args)}")
        
        lookup, if_set, if_unset = list(args) + [None]*(3 - len(args))

        value = self._evaluate_result(lookup, allow_unset=True)
        if type(value) is UNSET:
            if is_empty(if_unset):
                return UNSET
            return self._evaluate_result(if_unset)
        elif is_empty(if_set):
            return value
        else:
            return self._evaluate_result(if_set)            

    def func_GLOB(self, *args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(self.location)}: GLOB() expects 1 argument, got {len(args)}")
        pattern = self._evaluate_result(args[0])
        return sorted(glob.glob(pattern))

    def func_EXISTS(self, *args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(self.location)}: EXISTS() expects 1 argument, got {len(args)}")
        pattern = self._evaluate_result(args[0])
        return bool(glob.glob(pattern))

    def _evaluate_result(self, parse_result, allow_unset=False):
        # if result is already a constant, resolve it
        if type(parse_result) is not ParseResults:
            return self._resolve(parse_result)

        # if result is a unary or binary operator:
        if hasattr(parse_result, 'op1') and parse_result.op1:
            assert parse_result.op1 in self._UNARY_OPERATORS
            return self._UNARY_OPERATORS[parse_result.op1](parse_result[1])
        elif hasattr(parse_result, 'op2') and parse_result.op2:
            assert parse_result.op2 in self._BINARY_OPERATORS
            return self._BINARY_OPERATORS[parse_result.op2](parse_result[0], parse_result[2])

        # else lookup processing method based on name
        assert parse_result._name is not None
        method = parse_result._name
        if not hasattr(self, method):
            raise ParserError(f"{'.'.join(self.location)}: don't know how to deal with an element of type '{method}'")

        value = getattr(self, method)(*parse_result)

        if type(value) is UNSET and not allow_unset:
            raise SubstitutionError(f"{'.'.join(self.location)}: '{value.name}' is not defined")
        
        return value

    def evaluate(self, value, sublocation: List[str] = []):
        if type(value) is not str:
            return value

        loclen = len(self.location)
        self.location += sublocation

        try:
            if value.startswith("="):
                if value.startswith("=="):
                    return self._resolve(value[1:])
                else:
                    try:
                        parse_results = parse_string(value[1:])
                    except Exception as exc:
                        raise ParserError(f"{'.'.join(self.location)}: error parsing formula ({exc})")

                    return self._evaluate_result(parse_results)
            return self._resolve(value)
        finally:
            self.location = self.location[:loclen]
            

    def evaluate_dict(self, params: Dict[str, Any], 
                    corresponding_ns: typing.Optional[Dict[str, Any]] = None, 
                    defaults: Dict[str, Any] = {}, 
                    raise_substitution_errors: bool = True, verbose: bool =False):
        params = params.copy()
        for name, value in list(params.items()):
            if type(value) is not Unresolved:
                if verbose:
                    print(f"{name}: {value} ...")
                try:
                    new_value = self.evaluate(value, sublocation=[name])
                except SubstitutionError as err:
                    if raise_substitution_errors:
                        raise
                    new_value = Unresolved(errors=[err])
                if verbose:
                    print(f"{name}: {value} -> {new_value}")
                # UNSET return means delete or revert to default
                if new_value is UNSET:
                    if name in defaults:
                        params[name] = defaults[name]
                        if corresponding_ns:
                            corresponding_ns[name] = str(defaults[name])
                    else: 
                        del params[name]
                        if corresponding_ns and name in corresponding_ns:
                            del corresponding_ns[name]
                elif new_value is not value and new_value != value:
                    params[name] = new_value
                    if corresponding_ns:
                        corresponding_ns[name] = new_value
        return params

if __name__ == "__main__":
    pass
